import time
import uuid
import random

from flask_redis import FlaskRedis

# lua脚本，用来释放分布式锁, Redis执行lua脚本是原子性
# 保证 check and del 是原子性
LOCK_PREFIX = 'RLOCK_'
LUA_SCRIPT = "if redis.call('get', KEYS[1]) == KEYS[2] then return redis.call('del',KEYS[1]) else return 0 end"


class RedisRO(object):
    def __init__(self):
        self.clients = []

    def init_app(self, app):
        if app is not None:
            ro_count = app.config.get('REDIS_RO_COUNT', 0)
            if ro_count > 0:
                for i in range(0, ro_count):
                    c = FlaskRedis(app=app, config_prefix='REDIS_RO{}'.format(i + 1), decode_responses=True)
                    self.clients.append(c)

    @property
    def client(self):
        if len(self.clients) > 0:
            # 30%机会，让主节点承担查询操作
            # if random.randint(1, 10) in [6, 7, 8]:
            #     return redis_store
            return self.clients[random.randint(0, len(self.clients) - 1)]
        return redis_store


def get_lock(key, val='1', expire=10):
    return redis_store.set(LOCK_PREFIX + key, val, ex=expire, nx=True)


class RedisLock:
    """
    仅限需要进行原子操作的业务使用
    lock_name代表锁的范围， 比如lock_name=order_123
    代表对123这个订单加锁
    """

    def __init__(self, lock_name, interval_ms=10, timeout_ms=3000, expire=8, timeout=None):
        # timeout_ms 等待锁的超时时间(毫秒)
        # expire 锁的存活时间(秒), (在redis突然挂掉的时候,用于防止死锁现象)
        # 如果业务代码在锁的存活时间内还没执行完，就不能保证原子性了
        if timeout:
            timeout_ms = 1000 * timeout
        self.lock_name = lock_name
        self.timeout_ms = timeout_ms
        self.expire = expire
        if interval_ms > 200:
            interval_ms = 200
        if interval_ms < 0:
            interval_ms = 0
        self.interval_ms = interval_ms / 1000.0
        self.val = '{0}_{1}'.format(lock_name, str(uuid.uuid4()).replace('-', ''))

    def __enter__(self):
        return self.get()

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def get(self):
        # 获取锁, 阻塞
        end = self.get_utc_timestamp() + self.timeout_ms
        while True:
            if get_lock(self.lock_name, val=self.val, expire=self.expire):
                return True
            if self.get_utc_timestamp() > end:
                break
            if self.interval_ms > 0:
                time.sleep(self.interval_ms)
        return False

    def close(self):
        # 释放锁时，根据value判断，是不是我的锁，不能释放别人的锁
        redis_store.eval(LUA_SCRIPT, 2, LOCK_PREFIX + self.lock_name, self.val)

    @staticmethod
    def get_utc_timestamp():
        timestamp = time.time()
        return int(timestamp * 1000)


redis_store = FlaskRedis(decode_responses=True)
redis_ro_store = RedisRO()



# -*- coding: utf-8 -*-
import hashlib
import random
import string
import time
from abc import abstractmethod
from typing import Optional

from flask import request
from pydantic import BaseModel, Field

from easi_py_common.core.error import ServiceException


def __is_int(s):
    try:
        int(s)
        return True
    except Exception:
        pass
    return False


def __sign(app_key: str, app_secret: str, url: str, method: str, body_str: str, timestamp: str, nonce: str,
           salt: str = '&') -> str:
    signs = [app_key, app_secret, timestamp, nonce, method, url, body_str]
    signs = sorted(signs)
    sg = salt.join(signs)

    h = hashlib.md5()
    h.update(sg.encode("utf8"))
    sign = h.hexdigest()
    return sign.upper()


def sign(app_key: str, app_secret: str, url: str, method: str, body_str: str = '', salt: str = '&') -> (str, str, str):
    timestamp = str(int(time.time() * 1000))
    nonce = ''.join(random.sample(string.digits + string.ascii_letters, 32))
    sign = __sign(app_key, app_secret, url, method, body_str, timestamp, nonce, salt)

    return timestamp, nonce, sign


def valid(app_key: str, app_secret: str, url: str, method: str, body_str: str, timestamp: str, nonce: str,
          in_sign: str, salt: str = '&') -> bool:
    if not nonce or len(nonce) < 16:
        return False
    if not timestamp or not __is_int(timestamp):
        return False
    if not in_sign:
        return False

    generate_sign = __sign(app_key, app_secret, url, method, body_str, timestamp, nonce, salt)
    return generate_sign == str(in_sign).upper()


class AppInfo(BaseModel):
    app_key: str = Field(..., description='app_key')
    app_secret: str = Field(..., description='app_secret')
    self_info: Optional[object] = Field(description='self_info')


class AppInfoService:
    @abstractmethod
    def get_app_info(self, app_key: str) -> AppInfo: pass


class AppSignValidService:
    def __init__(self, app_info_service: AppInfoService, salt: str = '&',
                 header_app_key: str = 'app_key',
                 header_timestamp: str = 'timestamp',
                 header_nonce: str = 'nonce',
                 header_sign: str = 'sign'):

        self.app_info_service = app_info_service
        self.salt = salt
        self.header_app_key = header_app_key
        self.header_timestamp = header_timestamp
        self.header_nonce = header_nonce
        self.header_nonce = header_nonce
        self.header_sign = header_sign

    def __get_header_value(self, name):
        return request.headers[name] if name in request.headers else None

    def __get_url(self):
        url = request.full_path
        index = url.index("?")
        if index != -1:
            if len(url) - index == 1:
                url = url[0:index]
        return url

    def valid(self):
        app_key = self.__get_header_value(self.header_app_key)
        timestamp = self.__get_header_value(self.header_timestamp)
        nonce = self.__get_header_value(self.header_nonce)
        in_sign = self.__get_header_value(self.header_sign)
        request_data = request.get_data().decode('utf-8')
        url = self.__get_url()

        app_info = self.app_info_service.get_app_info(app_key)
        if not valid(app_key=app_info.app_key, app_secret=app_info.app_secret, url=url, method=request.method.upper(),
                     body_str=request_data, timestamp=timestamp, nonce=nonce, in_sign=in_sign, salt=self.salt):
            raise ServiceException(code=400, msg=u"invalid sign")

        return app_info.self_info



# -*- coding: utf-8 -*-
import inspect
from functools import wraps

from easi_py_common.client.rest_client import RestClient

service_endpoint = {}

HTTP_METHOD_GET = 'GET'
HTTP_METHOD_POST = 'POST'
HTTP_METHOD_PUT = 'PUT'
HTTP_METHOD_DELETE = 'DELETE'


def service_client(url, timeout=(5, 5), http_adapter=None, before_request=None):
    def wrapper(f):
        key = '{}.{}'.format(f.__module__, f.__name__)

        service_endpoint[key] = RestClient(url, timeout=timeout,
                                           http_adapter=http_adapter, before_request=before_request)

        @wraps(f)
        def wrapped(*args, **kwargs):
            return f(*args, **kwargs)

        return wrapped

    return wrapper


def service_mapping(uri, method):
    def wrapper(f):
        try:
            is_method = inspect.getargspec(f)[0][0] == 'self'
        except:
            is_method = False

        if not is_method:
            raise Exception('{} is invalid service_mapping', f.__name__)

        @wraps(f)
        def wrapped(*args, **kwargs):
            name = '{}.{}'.format(args[0].__class__.__module__, args[0].__class__.__name__)
            if name not in service_endpoint:
                raise Exception('{}.{} is not have config endpoint', f.__module__, f.__name__)

            _uri = uri.format(**kwargs)
            data = args[1] if len(args) > 1 else None
            if data is None:
                data = kwargs.get('data', None)
            client = service_endpoint[name]
            if method == HTTP_METHOD_GET:
                return client.get(_uri, params=data)
            elif method == HTTP_METHOD_POST:
                return client.post(_uri, json_data=data)
            elif method == HTTP_METHOD_PUT:
                return client.put(_uri, json_data=data)
            elif method == HTTP_METHOD_DELETE:
                return client.delete(_uri)
            else:
                raise Exception('{}.{} is invalid method', f.__module__, f.__name__)

        return wrapped

    return wrapper



import uuid

from flask import current_app
from itsdangerous import TimedJSONWebSignatureSerializer as Serializer, SignatureExpired

from easi_py_common.core.error import InvalidTokenException, TokenExpiredException
from easi_py_common.jwt.constants import JwtConstants


class Jwt:

    def __init__(self, secret_key: str = '', refresh_secret_key: str = ''):
        if not secret_key:
            secret_key = current_app.config['JWT_SECRET_KEY']
            refresh_secret_key = current_app.config['JWT_REFRESH_SECRET_KEY']
        self.secret_key = secret_key
        self.refresh_secret_key = refresh_secret_key

    def create_token(self, id: int, expires_in: int = JwtConstants.DEFAULT_EXPIRED_IN) -> str:
        s = self.__get_jw_serializer(self.secret_key, expires_in)
        return self.__create_token(s, id)

    def create_refresh_token(self, id: int, expires_in: int = JwtConstants.DEFAULT_REFRESH_EXPIRED_IN) -> str:
        s = self.__get_jw_serializer(self.refresh_secret_key, expires_in)
        return self.__create_token(s, id)

    def parse_token(self, token: str) -> int:
        s = self.__get_jw_serializer(self.secret_key)
        return self.__parse_token(s, token)

    def parse_refresh_token(self, refresh_token: str) -> int:
        s = self.__get_jw_serializer(self.refresh_secret_key)
        return self.__parse_token(s, refresh_token)

    def __create_token(self, s: Serializer, id: int):
        return s.dumps({'id': id, 'uid': uuid.uuid1().__str__()}).decode('ascii')

    def __parse_token(self, s: Serializer, token: str) -> int:
        try:
            data = s.loads(token)
        except SignatureExpired:
            raise TokenExpiredException()
        except Exception:
            raise InvalidTokenException()
        if 'id' not in data:
            raise InvalidTokenException()
        return int(data['id'])

    def __get_jw_serializer(self, secret_key: str, expires_in: int = 3600) -> Serializer:
        return Serializer(secret_key,
                          expires_in=expires_in,
                          algorithm_name=JwtConstants.SERIALIZER_ALGORITHM_NAME)



# -*- coding: utf-8 -*-
import random
from functools import partial

import sqlalchemy.orm as orm
from flask import current_app
from flask_sqlalchemy import SQLAlchemy as _BaseSQLAlchemy
from flask_sqlalchemy import get_state
from sqlalchemy.ext.declarative import declarative_base

MASTER_DB = 'master'
SLAVE_DB = 'slave'
FIXED_SLAVE_BINDS = 'fixed_slave'
ANL_SLAVE_BINDS = 'anl_slave'  # 统计分析库
DW_SLAVE_BINDS = 'dw_slave'  # dw分析库


class RoutingSession(orm.Session):

    def __init__(self, db, autocommit=False, autoflush=False, **options):
        self.app = db.get_app()
        self.db = db
        self._model_changes = {}

        # session 唯一 id
        # self.sid = get_random_string(length=8)
        orm.Session.__init__(
            self, autocommit=autocommit, autoflush=autoflush,
            bind=db.engine,
            binds=db.get_binds(self.app), **options)

    def get_bind(self, mapper=None, clause=None):

        # 被操作对象的包名
        full_class_name = ''
        if mapper is not None:
            full_class_name = "%s.%s" % (
                mapper.class_.__module__,
                mapper.class_.__name__
            )

        try:
            state = get_state(self.app)
        except (AssertionError, AttributeError, TypeError) as err:
            self.app.logger.info.warning(
                "cant get configuration. default bind. Error:" + err)
            return orm.Session.get_bind(self, mapper, clause)

        """
        If there are no binds configured, connect using the default
        SQLALCHEMY_DATABASE_URI
        """
        if state is None or not self.app.config['SQLALCHEMY_BINDS']:
            if not self.app.debug:
                self.app.logger.info("Connecting -> DEFAULT")
            return orm.Session.get_bind(self, mapper, clause)

        elif self._name:
            # self.app.logger.info("Connecting -> {}".format(self._name))
            return state.db.get_engine(self.app, bind=self._name)

        # 写入请求使用 master
        # 操作被改动过的对象，也使用 master
        elif self._flushing or self._model_changes.get(full_class_name):
            bind_key = getattr(mapper.class_, "__bind_key__", MASTER_DB)

            self._model_changes[full_class_name] = True
            if self.app.debug:
                action = 'WRITE' if self._flushing else 'READ'
                self.app.logger.info(
                    "Connecting -> %s MASTER( %s - %s )" % (
                        action,
                        mapper.class_,
                        bind_key
                    )
                )
            return state.db.get_engine(self.app, bind=bind_key)

        # 其他请求使用 slave
        else:
            bind_key = getattr(mapper.class_, "__bind_key__", SLAVE_DB)
            slave_bind_key = "%s_slave" % bind_key
            if slave_bind_key in self.app.config['SQLALCHEMY_BINDS']:
                bind_key = slave_bind_key

            if self.app.debug:
                self.app.logger.info(
                    "Connecting -> READ SLAVE( %s - %s )" % (
                        mapper.class_, bind_key
                    )
                )
            return state.db.get_engine(self.app, bind=bind_key)

    _name = None

    def using_bind(self, name):
        s = RoutingSession(self.db)
        vars(s).update(vars(self))
        s._name = name
        return s


class SQLAlchemy(_BaseSQLAlchemy):
    def apply_pool_defaults(self, app, options):
        super(SQLAlchemy, self).apply_pool_defaults(app, options)
        options["pool_pre_ping"] = True


class RouteSQLAlchemy(SQLAlchemy):
    """可读写分离路由的 SQLALCHEMY 对象"""

    def __init__(self, *args, **kwargs):
        SQLAlchemy.__init__(self, *args, **kwargs)
        self.session.using_bind = lambda s: self.session().using_bind(s)

    def create_scoped_session(self, options=None):
        if options is None:
            options = {}
        scopefunc = options.pop('scopefunc', None)
        # print "create_scoped_session - options = ", options
        return orm.scoped_session(
            partial(RoutingSession, self, **options), scopefunc=scopefunc
        )

    def apply_driver_hacks(self, app, info, options):
        """设置一些 create_engine 时的参数"""

        if "isolation_level" not in options and hasattr(app.config, "ISOLATION_LEVEL"):
            options["isolation_level"] = app.config["ISOLATION_LEVEL"]

        if hasattr(app.config, "ECHO_RAW_SQL"):
            options['echo'] = app.config["ECHO_RAW_SQL"]
            # print "update options - ", options

        return super(RouteSQLAlchemy, self).apply_driver_hacks(app, info, options)

    def get_engine(self, app=None, bind=None):
        # 从固定的slave里随机获取一个，如果配置了的话
        if bind == FIXED_SLAVE_BINDS and 'SQLALCHEMY_BINDS_FIXED_SLAVE_COUNT' in current_app.config and \
                current_app.config['SQLALCHEMY_BINDS_FIXED_SLAVE_COUNT'] > 0:
            # 随机选则
            fixed_bind = '{}{}'.format(FIXED_SLAVE_BINDS,
                                       random.randint(1, current_app.config['SQLALCHEMY_BINDS_FIXED_SLAVE_COUNT']))
            try:
                return super(RouteSQLAlchemy, self).get_engine(current_app, fixed_bind)
            except AssertionError as e:
                current_app.logger.error('get fixed engine error:{}'.format(e))

        if bind == FIXED_SLAVE_BINDS:
            bind = SLAVE_DB

        return super(RouteSQLAlchemy, self).get_engine(app, bind)

    @property
    def engine(self):
        return self.get_engine(bind=SLAVE_DB)

    @property
    def fixed_engine(self):
        return self.get_engine(bind=FIXED_SLAVE_BINDS)

    @property
    def anl_engine(self):
        return self.get_engine(bind=ANL_SLAVE_BINDS)

    @property
    def dw_engine(self):
        return self.get_engine(bind=DW_SLAVE_BINDS)


def slave_db():
    return db.session.using_bind(SLAVE_DB)


def master_db():
    return db.session.using_bind(MASTER_DB)


def using_master_session_query(model):
    if model is None:
        return None
    return master_db().query(model)


def using_fixed_slave_session_query(model):
    if model is None:
        return None
    return db.session.using_bind(FIXED_SLAVE_BINDS).query(model)


db: SQLAlchemy = RouteSQLAlchemy()
Model = declarative_base()



from easi_py_common.sqs.consumer import subscribe as sub


def subscribe(address,
              work_number=8,  # 协程数
              msg_lifetime=86400 * 3,  # 消息过期时间(xx秒内未消费直接丢弃)
              visibility_timeout_seconds=300):  # 消息超时时间，超过xx秒未执行完成，会走重试流程
    def wrapper(f):
        sub(address, f,
            work_number=work_number,
            msg_lifetime=msg_lifetime,
            visibility_timeout_seconds=visibility_timeout_seconds)
        return f

    return wrapper



from gevent.pool import Pool

pool = Pool(64)


class ProducerSession:
    buffer_size = 10

    def __init__(self, sqs_client, sign, app=None, cb=None):
        self.sqs_client = sqs_client
        self.sign = sign
        self.app = app
        self.cb = cb
        self.msgs = []

    def add(self, address, msg, key='', group_id='', delay_seconds=0, sign=None):
        if sign is None:
            address = self.sign + address
        else:
            address = sign + address

        self.msgs.append(dict(
            address=address,
            body=msg,
            key=key,
            group_id=group_id,
            delay_seconds=delay_seconds
        ))

    def flush(self):
        self.msgs = []

    def commit(self):
        if not self.msgs:
            return
        temp_msgs = self.msgs[:]
        self.msgs = []
        msg_group_map = {}

        for msg in temp_msgs:
            if msg.address not in msg_group_map:
                msg_group_map[msg.address] = []
            msg_group_map[msg.address].append(msg)
        for address, msg_list in msg_group_map.items():
            msg_buffer_list = []
            for m in msg_list:
                msg_buffer_list.append(m)
                if len(msg_buffer_list) % self.buffer_size == 0:
                    pool.spawn(self.sqs_client.send_message_batch, address, msg_buffer_list, cb=self.cb, app=self.app)
                    msg_buffer_list = []
            if msg_buffer_list:
                pool.spawn(self.sqs_client.send_message_batch, msg_buffer_list, cb=self.cb, app=self.app)
